﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.IO;
using System.Threading.Tasks;

namespace System.Net.Http.Test
{
    [TestClass]
    public class MultipartFormDataContentTest
    {
        [TestMethod]
        public void Ctor_NoParams_CorrectMediaType()
        {
            MultipartFormDataContent content = new MultipartFormDataContent();
            Assert.AreEqual("multipart/form-data", content.Headers.ContentType.MediaType);
            Assert.AreEqual(1, content.Headers.ContentType.Parameters.Count);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void Ctor_NullBoundary_Throw()
        {
            MultipartFormDataContent content = new MultipartFormDataContent(null);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void Ctor_EmptyBoundary_Throw()
        {
            MultipartFormDataContent content = new MultipartFormDataContent(String.Empty);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Add_NullContent_Throw()
        {
            MultipartFormDataContent content = new MultipartFormDataContent();
            content.Add(null);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void Add_NullName_Throw()
        {
            MultipartFormDataContent content = new MultipartFormDataContent();
            content.Add(new StringContent("Hello world"), null);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void Add_EmptyName_Throw()
        {
            MultipartFormDataContent content = new MultipartFormDataContent();
            content.Add(new StringContent("Hello world"), String.Empty);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void Add_NullFileName_Throw()
        {
            MultipartFormDataContent content = new MultipartFormDataContent();
            content.Add(new StringContent("Hello world"), "name", null);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void Add_EmptyFileName_Throw()
        {
            MultipartFormDataContent content = new MultipartFormDataContent();
            content.Add(new StringContent("Hello world"), "name", String.Empty);
        }

        [TestMethod]
        public void Serialize_EmptyList_Success()
        {
            MultipartFormDataContent content = new MultipartFormDataContent("test_boundary");
            MemoryStream output = new MemoryStream();
            content.CopyTo(output);

            output.Seek(0, SeekOrigin.Begin);
            string result = new StreamReader(output).ReadToEnd();

            long length;
            Assert.IsTrue(content.TryComputeLength(out length));
            Assert.AreEqual(result.Length, length);

            Assert.AreEqual("--test_boundary\r\n\r\n--test_boundary--\r\n", result);
        }

        [TestMethod]
        public void SerializeAsync_EmptyList_Success()
        {
            MultipartFormDataContent content = new MultipartFormDataContent("test_boundary");
            MemoryStream output = new MemoryStream();
            Task task = content.CopyToAsync(output);

            Assert.IsTrue(task.Wait(1000), "Test Hung");

            output.Seek(0, SeekOrigin.Begin);
            string result = new StreamReader(output).ReadToEnd();

            long length;
            Assert.IsTrue(content.TryComputeLength(out length));
            Assert.AreEqual(result.Length, length);

            Assert.AreEqual("--test_boundary\r\n\r\n--test_boundary--\r\n", result);
        }

        [TestMethod]
        public void Serialize_StringContent_Success()
        {
            MultipartFormDataContent content = new MultipartFormDataContent("test_boundary");
            content.Add(new StringContent("Hello World"));

            MemoryStream output = new MemoryStream();
            content.CopyTo(output);

            output.Seek(0, SeekOrigin.Begin);
            string result = new StreamReader(output).ReadToEnd();

            long length;
            Assert.IsTrue(content.TryComputeLength(out length));
            Assert.AreEqual(result.Length, length);

            Assert.AreEqual(
                "--test_boundary\r\nContent-Type: text/plain; charset=utf-8\r\n"
                + "Content-Disposition: form-data\r\n\r\nHello World\r\n--test_boundary--\r\n",
                result);
        }

        [TestMethod]
        public void SerializeAsync_StringContent_Success()
        {
            MultipartFormDataContent content = new MultipartFormDataContent("test_boundary");
            content.Add(new StringContent("Hello World"));

            MemoryStream output = new MemoryStream();
            Task task = content.CopyToAsync(output);

            Assert.IsTrue(task.Wait(1000), "Test Hung");

            output.Seek(0, SeekOrigin.Begin);
            string result = new StreamReader(output).ReadToEnd();

            long length;
            Assert.IsTrue(content.TryComputeLength(out length));
            Assert.AreEqual(result.Length, length);

            Assert.AreEqual(
                "--test_boundary\r\nContent-Type: text/plain; charset=utf-8\r\n"
                + "Content-Disposition: form-data\r\n\r\nHello World\r\n--test_boundary--\r\n",
                result);
        }

        [TestMethod]
        public void Serialize_NamedStringContent_Success()
        {
            MultipartFormDataContent content = new MultipartFormDataContent("test_boundary");
            content.Add(new StringContent("Hello World"), "test_name");

            MemoryStream output = new MemoryStream();
            content.CopyTo(output);

            output.Seek(0, SeekOrigin.Begin);
            string result = new StreamReader(output).ReadToEnd();

            long length;
            Assert.IsTrue(content.TryComputeLength(out length));
            Assert.AreEqual(result.Length, length);

            Assert.AreEqual(
                "--test_boundary\r\nContent-Type: text/plain; charset=utf-8\r\n"
                + "Content-Disposition: form-data; name=test_name\r\n\r\nHello World\r\n--test_boundary--\r\n",
                result);
        }

        [TestMethod]
        public void Serialize_FileNameStringContent_Success()
        {
            MultipartFormDataContent content = new MultipartFormDataContent("test_boundary");
            content.Add(new StringContent("Hello World"), "test_name", "test_file_name");

            MemoryStream output = new MemoryStream();
            content.CopyTo(output);

            output.Seek(0, SeekOrigin.Begin);
            string result = new StreamReader(output).ReadToEnd();

            long length;
            Assert.IsTrue(content.TryComputeLength(out length));
            Assert.AreEqual(result.Length, length);

            Assert.AreEqual(
                "--test_boundary\r\nContent-Type: text/plain; charset=utf-8\r\n"
                + "Content-Disposition: form-data; name=test_name; " 
                + "filename=test_file_name; filename*=utf-8\'\'test_file_name\r\n\r\n"
                + "Hello World\r\n--test_boundary--\r\n",
                result);
        }

        [TestMethod]
        public void Serialize_QuotedName_Success()
        {
            MultipartFormDataContent content = new MultipartFormDataContent("test_boundary");
            content.Add(new StringContent("Hello World"), "\"test name\"");

            MemoryStream output = new MemoryStream();
            content.CopyTo(output);

            output.Seek(0, SeekOrigin.Begin);
            string result = new StreamReader(output).ReadToEnd();

            long length;
            Assert.IsTrue(content.TryComputeLength(out length));
            Assert.AreEqual(result.Length, length);

            Assert.AreEqual(
                "--test_boundary\r\nContent-Type: text/plain; charset=utf-8\r\n"
                + "Content-Disposition: form-data; name=\"test name\"\r\n\r\nHello World\r\n--test_boundary--\r\n",
                result);
        }

        [TestMethod]
        public void Serialize_InvalidName_Encoded()
        {
            MultipartFormDataContent content = new MultipartFormDataContent("test_boundary");
            content.Add(new StringContent("Hello World"), "testク\r\n namé");

            MemoryStream output = new MemoryStream();
            content.CopyTo(output);

            output.Seek(0, SeekOrigin.Begin);
            string result = new StreamReader(output).ReadToEnd();

            long length;
            Assert.IsTrue(content.TryComputeLength(out length));
            Assert.AreEqual(result.Length, length);

            Assert.AreEqual(
                "--test_boundary\r\nContent-Type: text/plain; charset=utf-8\r\n"
                + "Content-Disposition: form-data; name=\"=?utf-8?B?dGVzdOOCrw0KIG5hbcOp?=\""
                + "\r\n\r\nHello World\r\n--test_boundary--\r\n",
                result);
        }

        [TestMethod]
        public void Serialize_InvalidQuotedName_Encoded()
        {
            MultipartFormDataContent content = new MultipartFormDataContent("test_boundary");
            content.Add(new StringContent("Hello World"), "\"testク\r\n namé\"");

            MemoryStream output = new MemoryStream();
            content.CopyTo(output);

            output.Seek(0, SeekOrigin.Begin);
            string result = new StreamReader(output).ReadToEnd();

            long length;
            Assert.IsTrue(content.TryComputeLength(out length));
            Assert.AreEqual(result.Length, length);

            Assert.AreEqual(
                "--test_boundary\r\nContent-Type: text/plain; charset=utf-8\r\n"
                + "Content-Disposition: form-data; name=\"=?utf-8?B?dGVzdOOCrw0KIG5hbcOp?=\""
                + "\r\n\r\nHello World\r\n--test_boundary--\r\n",
                result);
        }

        [TestMethod]
        public void Serialize_InvalidNamedFileName_Encoded()
        {
            MultipartFormDataContent content = new MultipartFormDataContent("test_boundary");
            content.Add(new StringContent("Hello World"), "testク\r\n namé", "fileク\r\n namé");

            MemoryStream output = new MemoryStream();
            content.CopyTo(output);

            output.Seek(0, SeekOrigin.Begin);
            string result = new StreamReader(output).ReadToEnd();

            long length;
            Assert.IsTrue(content.TryComputeLength(out length));
            Assert.AreEqual(result.Length, length);

            Assert.AreEqual(
                "--test_boundary\r\nContent-Type: text/plain; charset=utf-8\r\n"
                + "Content-Disposition: form-data; name=\"=?utf-8?B?dGVzdOOCrw0KIG5hbcOp?=\";"
                + " filename=\"=?utf-8?B?ZmlsZeOCrw0KIG5hbcOp?=\"; filename*=utf-8\'\'file%E3%82%AF%0D%0A%20nam%C3%A9" 
                + "\r\n\r\nHello World\r\n--test_boundary--\r\n",
                result);
        }
    }
}
